#!/usr/bin/python

"""The HtmlHandler class does the actual filtering of the HTML: it
spell-checks every Hebrew word and adds the appropriate markup."""

# This file is released under the terms of the GPL
# Copyright (C) Mooffie <mooffie@typo.co.il>

import cfg
import sys
import re
import htmlparser
import hspell
import string

# Instantiate a global speller object:
speller = hspell.Hspell()

# Elements whose CDATA we should not spell-check:
excluded_elts = ('script', 'textarea', 'style', 'title', 'select')

CHRST_WIN1255	     = 1
CHRST_WIN1255_VISUAL = 2
CHRST_UTF8	     = 3
CHRST_NOHEB	     = 4 # e.g. iso-8859-1

def get_charset(s, fallback=CHRST_WIN1255):
    """Gets a string of the form 'content-type: text/html; charset=utf-8' and
       parses the 'charset' attribute"""
    map = ( ('utf-8',	     CHRST_UTF8),
	    ('windows-1255', CHRST_WIN1255),
	    ('iso-8859-8-i', CHRST_WIN1255),
	    ('iso-8859-8i',  CHRST_WIN1255),
	    ('iso-8859-8',   CHRST_WIN1255_VISUAL),
	    ('logical',	     CHRST_WIN1255),
	    ('visual',	     CHRST_WIN1255_VISUAL) )
    s = s.lower()
    for ident, charset in map:
	if s.find(ident) != -1:
	    return charset
    if s.find("charset") != -1:
	return CHRST_NOHEB
    else:
	return fallback

def utf8_to_win1255(s):
    try:
	return unicode(s, 'utf-8').encode('iso-8859-8')
    except UnicodeError:
	return ""

def win1255_to_utf8(s):
    return unicode(s, 'iso-8859-8').encode('utf-8')

def revstr(s):
    lst = list(s)
    lst.reverse()
    return string.join(lst, "")

def escape_html(s):
    map = ( ("&", "&amp;"),
	    ('"', "&quot;"),
	    ("<", "&lt;"),
	    (">", "&gt;") )
    for old, new in map:
	s = string.replace(s, old, new)
    return s

# Regular expressions used in this module:
alpha = r'a-zA-Z\x80-\xFF'
alpha_q = r'a-zA-Z\x80-\xFF"' "'"
niqud_ent = r'&(?:#145\d|#146\d|quot);'
maqaf_win1255 = "\xCE"
maqaf_utf8 = "\xD6\xBE"
heb_alpha_re = re.compile(r'[\x80-\xFF]')
niqud_win1255_re = re.compile(r'[\xC0-\xD8\xFD\xFE]')
niqud_utf8_re = re.compile(r'(?:\xD6[\xB0-\xFF]|\xD7[\x00-\x83]|\xD7[\xB0-\xB4]|\xE2\x80[\x8E\x8F])')
# the following regexp determines what constitutes a word
split_words_re = re.compile('(['+alpha+'](?:['+alpha_q+']|'+niqud_ent+')*['+alpha+"']|["+alpha+'])')

class HtmlHandler(htmlparser.HtmlContentHandler):

    def __init__(self, out, default_charset):
	self.out = out
	self.charset = default_charset
	self.in_excluded_elt = 0

    def dtd(self, s):
	self.out.write(s)

    def startElement(self, name, tag_data):
	self.out.write(tag_data)
	if name == "meta":
	    # If you ever change the following line, make sure you
	    # capture the non-standard <META charset="..."> too.
	    if tag_data.lower().find("charset") != -1:
		self.charset = get_charset(tag_data)
	if name in excluded_elts:
	    self.in_excluded_elt = 1
	
    def endElement(self, name, tag_data):
	self.out.write(tag_data)
	if name in excluded_elts:
	    self.in_excluded_elt = 0
	
    def characters(self, s):
	if self.in_excluded_elt or self.charset == CHRST_NOHEB \
		or not heb_alpha_re.search(s): \
	    self.out.write(s)
	else:
	    self.out.write(self.spellcheck(s))
		
    def comment(self, s):
	self.out.write(s)

    def processingInstruction(self, s):
	self.out.write(s)

    def spellcheck(self, s):
	charset = self.charset
	# replace maqaf with its entity equivalent so that
	# we can split on it.
	if charset == CHRST_UTF8:
	    s = string.replace(s, maqaf_utf8, "&#1470;")
	else:
	    s = string.replace(s, maqaf_win1255, "&#1470;")

	speller.lock()
	ret = ""
	for word in split_words_re.split(s):
	    # spellcheck a word only if it contains some hebrew chars
	    # and does NOT contain some niqud.
	    if heb_alpha_re.search(word) \
		    and (not "&" in word) \
		    and ((charset != CHRST_UTF8 and not niqud_win1255_re.search(word)) or \
			 (charset == CHRST_UTF8 and not niqud_utf8_re.search(word))) \
		    :
		if charset == CHRST_UTF8:
		    win1255word = utf8_to_win1255(word)
		elif charset == CHRST_WIN1255_VISUAL:
		    win1255word = revstr(word)
		else:
		    win1255word = word

		# some degenerated web-boards mangle quotes...
		win1255word = string.replace(win1255word, "''", '"')

		splret = speller.check_word(win1255word)
		
		if splret and "'" in win1255word:
		    # maybe the word is reported as a misspelling just
		    # because hspell doesn't support ' that functions as "
		    if not speller.check_word(string.replace(win1255word, "'", '"')):
			# yes, that's the case. word is actually correct.
			splret = {}

		if not splret:
		    # word is correct
		    ret += word
		else:
		    # word is misspelled
		    suggestions = string.join(splret['suggestions'], ", ")
		    if not suggestions:
			suggestions = "(no suggestions)"
		    #suggestions = "{" + win1255word + "} " + suggestions
		    if splret['hint']:
			suggestions += " |" + splret['hint']
		    suggestions = escape_html(suggestions)
		    if charset == CHRST_UTF8:
			suggestions = win1255_to_utf8(suggestions)
		    ret += cfg.misspelled_word_template % (suggestions, word)
	    else:
		ret += word
	speller.unlock()
	return ret

# When started as a unix filter:
if __name__ == '__main__':
    default_charset = CHRST_WIN1255
    parser = htmlparser.HtmlParser(HtmlHandler(sys.stdout, default_charset))
    if len(sys.argv) != 1:
	f = file(sys.argv[1])
    else:
	f = sys.stdin
    while 1:
	# We must not break CDATA words, so we use readline()
	data = f.readline()
	if not data: break
	parser.parseString(data)


